# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals
import re

from ..utils import dictify
from .base import BaseFilterBackend


class PlainFilterBackend(BaseFilterBackend):
    """
    Filter backend for filtering plain Python iterables.

    .. warning::
        The filter backend does filtering inside a regular loop
        by comparing attributes of individual objects within iterable.
        As a result, this is **NOT** efficient method for filtering
        any large amounts of data. In those cases, it would probably
        be better to find more appropriate and efficient way to filter data.
    """

    name = "plain"
    enforce_same_models = False
    supported_lookups = {
        "contains",
        "day",
        "endswith",
        "exact",
        "gt",
        "gte",
        "hour",
        "icontains",
        "iendswith",
        "iexact",
        "iin",
        "in",
        "iregex",
        "isnull",
        "istartswith",
        "lt",
        "lte",
        "minute",
        "month",
        "range",
        "regex",
        "second",
        "startswith",
        "week_day",
        "year",
    }

    def empty(self):
        """
        Get empty queryset
        """
        return []

    def get_model(self):
        """
        Get the model from the given queryset

        Since there is no specific model for filtering Python lists,
        this simply returns ``object``
        """
        return object

    def filter_by_specs(self, queryset):
        """
        Filter queryset by applying all filter specifications

        The filtering is done by calling manually loping over all
        items in the iterable and comparing inner attributes with the
        filter specification.
        """
        if not self.regular_specs:
            return queryset

        return list(filter(self._filter_callable, queryset))

    def _filter_callable(self, item):
        return all(self._filter_by_spec(item, spec) for spec in self.regular_specs)

    def _filter_by_spec(self, item, spec):
        filtered = self._filter_by_spec_and_value(item, spec.components, spec)
        if spec.is_negated:
            return not filtered
        return filtered

    def _filter_by_spec_and_value(self, item, components, spec):
        if not components and not isinstance(item, (list, tuple)):
            comparator = getattr(self, "_compare_{}".format(spec.lookup))
            try:
                return comparator(item, spec)
            except Exception:
                return True

        if isinstance(item, (list, tuple)):
            return any(
                self._filter_by_spec_and_value(i, components, spec) for i in item
            )

        if not isinstance(item, dict):
            item = dictify(item)

        return self._filter_by_spec_and_value(
            item.get(components[0], {}), components[1:], spec
        )

    def _compare_contains(self, value, spec):
        return spec.value in value

    def _compare_day(self, value, spec):
        return value.day == spec.value

    def _compare_endswith(self, value, spec):
        return value.endswith(spec.value)

    def _compare_exact(self, value, spec):
        return value == spec.value

    def _compare_gt(self, value, spec):
        return value > spec.value

    def _compare_gte(self, value, spec):
        return value >= spec.value

    def _compare_hour(self, value, spec):
        return value.hour == spec.value

    def _compare_icontains(self, value, spec):
        return spec.value.lower() in value.lower()

    def _compare_iendswith(self, value, spec):
        return value.lower().endswith(spec.value.lower())

    def _compare_iexact(self, value, spec):
        return value.lower() == spec.value.lower()

    def _compare_in(self, value, spec):
        return value in spec.value

    def _compare_iin(self, value, spec):
        return value.lower() in [i.lower() for i in spec.value]

    def _compare_iregex(self, value, spec):
        return bool(re.match(spec.value, value, re.IGNORECASE))

    def _compare_isnull(self, value, spec):
        if spec.value:
            return value is None
        else:
            return value is not None

    def _compare_istartswith(self, value, spec):
        return value.lower().startswith(spec.value.lower())

    def _compare_lt(self, value, spec):
        return value < spec.value

    def _compare_lte(self, value, spec):
        return value <= spec.value

    def _compare_minute(self, value, spec):
        return value.minute == spec.value

    def _compare_month(self, value, spec):
        return value.month == spec.value

    def _compare_range(self, value, spec):
        return spec.value[0] <= value <= spec.value[1]

    def _compare_regex(self, value, spec):
        return bool(re.match(spec.value, value))

    def _compare_second(self, value, spec):
        return value.second == spec.value

    def _compare_startswith(self, value, spec):
        return value.startswith(spec.value)

    def _compare_week_day(self, value, spec):
        # expected 1-Sunday and 7-Saturday
        return ((value.weekday() + 1) % 7) + 1 == spec.value

    def _compare_year(self, value, spec):
        return value.year == spec.value



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

from django.core.exceptions import FieldDoesNotExist
from django.db.models.constants import LOOKUP_SEP

from ..utils import suppress
from .base import BaseFilterBackend


class DjangoFilterBackend(BaseFilterBackend):
    """
    Filter backend for filtering Django querysets.

    .. warning::
        The filter backend can **ONLY** filter Django's ``QuerySet``.
        Passing any other datatype for filtering will kill happy bunnies
        under rainbow.
    """

    name = "django"
    supported_lookups = {
        "contains",
        "day",
        "endswith",
        "exact",
        "gt",
        "gte",
        "hour",
        "icontains",
        "iendswith",
        "iexact",
        "in",
        "iregex",
        "isnull",
        "istartswith",
        "lt",
        "lte",
        "minute",
        "month",
        "range",
        "regex",
        "second",
        "startswith",
        "week_day",
        "year",
    }

    def empty(self):
        """
        Get empty queryset
        """
        return self.queryset.none()

    def get_model(self):
        """
        Get the model from the given queryset
        """
        return self.queryset.model

    @property
    def includes(self):
        """
        Property which gets list of non-negated filters

        By combining all non-negated filters we can optimize filtering by
        calling ``QuerySet.filter`` once rather then calling it for each
        filter specification.
        """
        return filter(lambda i: not i.is_negated, self.regular_specs)

    @property
    def excludes(self):
        """
        Property which gets list of negated filters

        By combining all negated filters we can optimize filtering by
        calling ``QuerySet.exclude`` once rather then calling it for each
        filter specification.
        """
        return filter(lambda i: i.is_negated, self.regular_specs)

    def _prepare_spec(self, spec):
        return "{}{}{}".format(
            LOOKUP_SEP.join(spec.components), LOOKUP_SEP, spec.lookup
        )

    def filter_by_specs(self, queryset):
        """
        Filter queryset by applying all filter specifications

        The filtering is done by calling ``QuerySet.filter`` and
        ``QuerySet.exclude`` as appropriate.
        """
        include = {self._prepare_spec(i): i.value for i in self.includes}
        exclude = {self._prepare_spec(i): i.value for i in self.excludes}

        if include:
            queryset = queryset.filter(**include)

        # Plain queryset.exclude(**exclude) would cause exclusion of ALL
        # negative-matching objects. I.e. x!=1&y!=2 is equivalent
        # to "NOT (x = 1 AND y = 2)" SQL, which is not an intuitive behavior.
        # We chain .exclude to achieve "NOT (x = 1) AND NOT (y = 2)" instead.
        for lookup, value in exclude.items():
            queryset = queryset.exclude(**{lookup: value})

        to_many = self._is_any_to_many()
        return queryset.distinct() if to_many and (include or exclude) else queryset

    def _is_any_to_many(self):
        return any(
            self._is_to_many(self.model, i.components) for i in self.regular_specs
        )

    def _is_to_many(self, model, components):
        if not components:
            return False

        with suppress(FieldDoesNotExist):
            f = model._meta.get_field(components[0])
            return (
                f.one_to_many
                or f.many_to_many
                or self._is_to_many(f.related_model, components[1:])
            )



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals
import abc
import re
from collections import defaultdict
from copy import deepcopy

import six
from cached_property import cached_property
from django.core.exceptions import ValidationError
from django.core.validators import RegexValidator
from django.db.models.constants import LOOKUP_SEP
from django.http import QueryDict

from ..backends.django import DjangoFilterBackend
from ..constants import StrictMode
from ..exceptions import Empty, SkipFilter
from ..filters import BaseFilter
from ..utils import LookupConfig


__all__ = ["FilterSet", "FilterSetOptions", "ModelFilterSetOptions"]

LOOKUP_RE = re.compile(
    r"^(?:[^\d\W]\w*)(?:{}?[^\d\W]\w*)*(?:!)?$" r"".format(LOOKUP_SEP), re.IGNORECASE
)


class FilterKeyValidator(RegexValidator):
    """
    Custom regex validator for validating the querystring filter
    is of correct syntax::

        name[__<relation>]*[__<lookup_method>][!]
    """

    regex = LOOKUP_RE
    message = (
        "Filter key is of invalid format. "
        "It must be `name[__<relation>]*[__<lookup_method>][!]`."
    )


filter_key_validator = FilterKeyValidator()


class FilterSetOptions(object):
    """
    Base class for handling options passed to :class:`.FilterSet`
    via ``Meta`` attribute.
    """

    def __init__(self, options=None):
        pass


class FilterSetMeta(type(BaseFilter)):
    """
    Metaclass for creating :class:`.FilterSet` classes.

    Its primary job is to do:

    * collect all declared filters in all bases
      and set them as ``_declared_filters`` on created
      :class:`.FilterSet` class.
    * instantiate ``Meta`` by using ``filter_options_class`` attribute
    """

    def __new__(cls, name, bases, attrs):
        try:
            parents = [b for b in bases if issubclass(b, FilterSet)]
        except NameError:
            parents = False

        new_class = super(FilterSetMeta, cls).__new__(cls, name, bases, attrs)

        # creating FilterSet itself
        if not parents:
            return new_class

        filters = {}
        for base in [vars(base) for base in bases] + [attrs]:
            filters.update({k: v for k, v in base.items() if isinstance(v, BaseFilter)})

        new_class._declared_filters = filters
        new_class.Meta = new_class.filter_options_class(
            getattr(new_class, "Meta", None)
        )

        return new_class


class FilterSet(six.with_metaclass(FilterSetMeta, BaseFilter)):
    """
    Main user-facing classes to use filtersets.

    It primarily does:

    * takes queryset to filter
    * takes querystring data which will be used to filter
      given queryset
    * from the querystring, it constructs a list of :class:`.LookupConfig`
    * loops over the created configs and attempts to get
      :class:`.FilterSpec` for each
    * in the process, it delegates the job of constructing spec
      to child filters when any match is found between filter
      defined on the filter and lookup name in the config

    Parameters
    ----------
    data : QueryDict, optional
        ``QueryDict`` of querystring data.
        Only optional when :class:`.FilterSet` is used as a nested filter
        within another :class:`.FilterSet`.
    queryset : iterable, optional
        Can be any iterable as supported by the filter backend.
        Only optional when :class:`.FilterSet` is used as a nested filter
        within another :class:`.FilterSet`.
    context : dict, optional
        Context for filtering. This is passed to filtering backend.
        Usually this would consist of passing ``request`` and ``view``
        object from the Django view.
    strict_mode : str, optional
        Strict mode how :class:`.FilterSet` should behave when any validation
        fails. See :class:`url_filter.constants.StrictMode` doc for more information.
        Default is ``empty``.
    """

    filter_backend_class = DjangoFilterBackend
    """
    Class to be used as filter backend. By default
    :class:`.DjangoFilterBackend` is used.
    """
    filter_options_class = FilterSetOptions
    """
    Class to be used to construct ``Meta`` during
    :class:`.FilterSet` class creation time in its metaclass.
    """
    default_strict_mode = StrictMode.empty
    """
    Default strict mode which should be used when one is not
    provided in initialization.
    """

    def __init__(
        self, data=None, queryset=None, context=None, strict_mode=None, *args, **kwargs
    ):
        super(FilterSet, self).__init__(*args, **kwargs)
        self.data = data
        self.queryset = queryset
        self.context = context or {}
        self.strict_mode = strict_mode or self.default_strict_mode

    def repr(self, prefix=""):
        """
        Custom representation of the filterset


        Parameters
        ----------
        prefix : str
            Prefix with which each line of the representation should
            be prefixed with. This allows to recursively get the
            representation of all descendants with correct indentation
            (children are indented compared to parent)
        """
        header = "{name}({source})".format(
            name=self.__class__.__name__,
            source='source="{}"'.format(self.source) if self.is_bound else "",
        )
        lines = [header] + [
            "{prefix}{key} = {value}".format(
                prefix=prefix + "  ", key=k, value=v.repr(prefix=prefix + "  ")
            )
            for k, v in sorted(self.filters.items())
        ]
        return "\n".join(lines)

    def get_filters(self):
        """
        Get all filters defined in this filterset.

        By default only declared filters are returned however
        this method is meant to be used as a hook in subclasses
        in order to enhance functionality such as automatically
        adding filters from model fields.
        """
        return deepcopy(self._declared_filters)

    @cached_property
    def filters(self):
        """
        Cached property for accessing filters available in this filteset.
        In addition to getting filters via :meth:`.get_filters`,
        this property binds all filters to the filterset by using
        :meth:`.BaseFilter.bind`.

        See Also
        --------
        get_filters
        :meth:`.BaseFilter.bind`
        """
        filters = self.get_filters()

        for name, _filter in filters.items():
            _filter.bind(name, self)

        return filters

    @cached_property
    def default_filter(self):
        """
        Cached property for looking up default filter.
        Default filter is a filter which is defined with ``is_default=True``.
        Useful when lookup config references nested filter without
        specifying which field to filter. In that case default filter
        will be used.
        """
        return next(
            iter(
                filter(lambda i: getattr(i, "is_default", False), self.filters.values())
            ),
            None,
        )

    def validate_key(self, key):
        """
        Validate that :class:`.LookupConfig` key is correct.

        This is the key as provided in the querystring.
        Currently key is validated against a regex expression.

        Useful to filter out invalid filter querystring pairs
        since not whole querystring is not dedicated for filter
        purposes but could contain other information such as
        pagination information. In that case if the key is
        invalid key for filtering, we can simply ignore it
        without wasting time trying to get filter specification
        for it.

        Parameters
        ----------
        key : str
            Key as provided in the querystring
        """
        filter_key_validator(key)

    def get_filter_backend(self):
        """
        Get instantiated filter backend class.

        This backend is then used to actually filter queryset.
        """
        return self.filter_backend_class(queryset=self.queryset, context=self.context)

    @cached_property
    def filter_backend(self):
        """
        Property for getting instantiated filter backend.

        Primarily useful when accessing filter_backend outside
        of the filterset such as leaf filters or integration
        layers since backend has useful information for both of
        those examples.
        """
        assert self.data is not None, (
            "Filter backend can only be used when data is provided " "to filterset."
        )
        return self.get_filter_backend()

    def filter(self):
        """
        Main method which should be used on root :class:`.FilterSet`
        to filter queryset.

        This method:

        * asserts that filtering is being done on root :class:`.FilterSet`
          and that all necessary data is provided
        * creates :class:`.LookupConfig` from the provided data (querystring)
        * loops over all configs and attempts to get :class:`.FilterSet`
          for all of them
        * instantiates filter backend
        * uses the created filter specs to filter queryset by using specs

        Returns
        -------
        querystring
            Filtered queryset
        """
        assert self.root is self, "``filter`` can only be called on root ``FilterSet``."
        assert self.queryset is not None, "``queryset`` was not passed for filtering."
        assert isinstance(
            self.data, QueryDict
        ), "``data`` should be an instance of QueryDict."

        try:
            specs = self.get_specs()
        except Empty:
            return self.filter_backend.empty()
        else:
            self.filter_backend.bind(specs)
            return self.filter_backend.filter()

    def get_specs(self):
        """
        Get list of :class:`.FilterSpec` for the given querystring data.

        This function does:

        * unpacks the querystring data to a list of :class:`.LookupConfig`
        * loops through all configs and uses appropriate children
          filters to generate list of :class:`.FilterSpec`
        * if any validations fails while generating specs,
          all errors are collected and depending on ``strict_mode``
          it re-raises the errors or ignores them.

        Returns
        -------
        list
            List of :class:`.FilterSpec`
        """
        configs = self._generate_lookup_configs()
        specs = []
        errors = defaultdict(list)

        for data in configs:
            try:
                self.validate_key(data.key)
            except ValidationError:
                continue

            try:
                specs.append(self.get_spec(data))
            except SkipFilter:
                pass
            except ValidationError as e:
                errors[data.key].extend(
                    getattr(e, "error_list", [getattr(e, "message", "")])
                )

        if errors:
            if self.strict_mode == StrictMode.fail:
                raise ValidationError(dict(errors))
            elif self.strict_mode == StrictMode.empty:
                raise Empty

        return specs

    def get_spec(self, config):
        """
        Get :class:`.FilterSpec` for the given :class:`.LookupConfig`.

        If the config is non leaf config (it has more nested fields),
        then the appropriate matching child filter is used
        to get the spec.
        If the config however is a leaf config, then ``default_filter``
        is used to get the spec, when available, and if not,
        this filter is skipped.

        Parameters
        ----------
        config : LookupConfig
            Config for which to generate :class:`.FilterSpec`

        Returns
        -------
        FilterSpec
            Individual filter spec
        """
        if isinstance(config.data, dict):
            name, value = config.name, config.value
        else:
            if self.default_filter is None:
                raise SkipFilter
            name = self.default_filter.source
            value = LookupConfig(config.key, config.data)

        if name not in self.filters:
            if self.default_filter and self is not self.root:
                # if name is not found as a filter, there is a possibility
                # it is a lookup made on the default filter of this filterset
                # in which case we try to get that spec directly from the child
                # however that is only allowed on nested filtersets
                # since on root filterset filter must be specified
                return self.default_filter.get_spec(config)
            else:
                raise SkipFilter

        return self.filters[name].get_spec(value)

    def _generate_lookup_configs(self):
        """
        Generate ``LookupConfig``s for all data in querystring data.
        """
        for key, values in self.data.lists():
            for value in values:
                yield LookupConfig(
                    key,
                    six.moves.reduce(
                        lambda a, b: {b: a},
                        (key.replace("!", "").split(LOOKUP_SEP) + [value])[::-1],
                    ),
                )


class ModelFilterSetOptions(FilterSetOptions):
    """
    Custom options for :class:`.FilterSet` used for model-generated filtersets.

    Attributes
    ----------
    model : Model
        Model class from which :class:`.FilterSet` will
        extract necessary filters.
    fields : None, list, optional
        Specific model fields for which filters
        should be created for.
        By default it is ``None`` in which case for all
        fields filters will be created for.
    exclude : list, optional
        Specific model fields for which filters
        should **not** be created for.
    allow_related : bool, optional
        Whether related/nested fields should be allowed
        when model fields are automatically determined
        (e.g. when explicit :attr:`.fields` is not provided).
    extra_kwargs : dict, optional
        Additional kwargs to be given to auto-generated individual filters
    """

    def __init__(self, options=None):
        super(ModelFilterSetOptions, self).__init__(options)
        self.model = getattr(options, "model", None)
        self.fields = getattr(options, "fields", None)
        self.exclude = getattr(options, "exclude", [])
        self.extra_kwargs = getattr(options, "extra_kwargs", {})
        self.allow_related = getattr(options, "allow_related", True)


class BaseModelFilterSet(FilterSet):
    """
    Base :class:`.FilterSet` for model-generated filtersets.

    The filterset can be configured via ``Meta`` class attribute,
    very much like how Django's ``ModelForm`` is configured.
    """

    filter_options_class = ModelFilterSetOptions

    def get_filters(self):
        """
        Get all filters defined in this filterset by introspecting
        the given model in ``Meta.model``.
        """
        filters = super(BaseModelFilterSet, self).get_filters()

        assert self.Meta.model, (
            "{name}.Meta.model is missing. Please specify the model "
            "in order to use {name}."
            "".format(name=self.__class__.__name__)
        )

        if self.Meta.fields in [None, "__all__"]:
            self.Meta.fields = self._get_model_field_names()

        state = self._build_state()

        for name in self.Meta.fields:
            if name in self.Meta.exclude or name in filters:
                continue

            try:
                _filter = self._build_filter(name, state)

            except SkipFilter:
                continue

            else:
                if _filter is not None:
                    filters[name] = _filter

        return filters

    @abc.abstractmethod
    def _get_model_field_names(self):
        """
        Get a list of all model fields.

        This is used when ``Meta.fields`` is ``None``
        in which case this method returns all model field names.

        .. note::
            This method is an abstract method and must be implemented
            in subclasses.
        """

    @abc.abstractmethod
    def _build_filter(self, name, state):
        """
        Build a filter for the field within the model by its name.

        .. note::
            This method is an abstract method and must be implemented
            in subclasses.

        Parameters
        ----------
        name : str
            Name of the field for which to build the filter within the ``Meta.model``
        state
            State of the model as returned by :meth:`._build_state`.
            Since state is computed outside of the loop which builds
            filters, state can be useful to store information outside
            of the loop so that it can be reused for all filters.
        """

    def _build_filterset(self, name, field_name, meta_attrs, base):
        """
        Helper method for building child filtersets.

        Parameters
        ----------
        name : str
            Name of the filterset to build. The returned class
            will use the name as a prefix in the class name.
        field_name : str
            Field name as filterset will be added to parent
            filterset. It will be used to extract ``extra_kwargs``.
        meta_attrs : dict
            Attributes to use for the ``Meta``.
        base : type
            Class to use as a base class for the filterset.
        """
        meta_attrs.update({"extra_kwargs": self._get_filter_extra_kwargs(field_name)})
        meta = type(str("Meta"), (object,), meta_attrs)

        filterset = type(
            str("{}FilterSet".format(name)),
            (base,),
            {"Meta": meta, "__module__": self.__module__},
        )

        return filterset()

    def _build_state(self):
        """
        Hook function to build state to be used while building all the filters.
        Useful to compute common data between all filters such as some
        data about model so that the computation can be avoided while
        building individual filters.
        """

    def _get_filter_extra_kwargs(self, name):
        """
        Get additional kwargs for individual filters
        as defined in Meta
        """
        return self.Meta.extra_kwargs.get(name, {})



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals
import inspect
from contextlib import contextmanager


class FilterSpec(object):
    """
    Class for describing filter specification.

    The main job of the :class:`.FilterSet` is to parse
    the submitted lookups into a list of filter specs.
    A list of these specs is then used by the filter backend
    to actually filter given queryset. That's what :class:`.FilterSpec`
    provides - a way to portably define filter specification
    to be used by a filter backend.

    The reason why filtering is decoupled from the :class:`.FilterSet`
    is because this allows to implement filter backends
    not related to Django.

    Attributes
    ----------
    components : list
        A list of strings which are names of the keys/attributes
        to be used in filtering of the queryset.
        For example lookup config with key
        ``user__profile__email`` will have components of
        ``['user', 'profile', 'email'].
    lookup : str
        Name of the lookup how final key/attribute from
        :attr:`.components` should be compared.
        For example lookup config with key
        ``user__profile__email__contains`` will have a lookup
        ``contains``.
    value
        Value of the filter.
    is_negated : bool, optional
        Whether this filter should be negated.
        By default its ``False``.
    filter_callable : func, optional
        Callable which should be used for filtering this
        filter spec. This is primaliry meant to be used
        by :class:`.CallableFilter`.
    """

    def __init__(
        self, components, lookup, value, is_negated=False, filter_callable=None
    ):
        self.components = components
        self.lookup = lookup
        self.value = value
        self.is_negated = is_negated
        self.filter_callable = filter_callable

    @property
    def is_callable(self):
        """
        Property for getting whether this filter specification is for a custom
        filter callable
        """
        return self.filter_callable is not None

    def __repr__(self):
        if self.is_callable:
            callable_repr = " via {}.{}".format(
                self.filter_callable.__self__.__class__.__name__,
                self.filter_callable.__name__,
            )
        else:
            callable_repr = ""

        return "<{name} {components} {negated}{lookup} {value!r}{callable}>".format(
            name=self.__class__.__name__,
            components=".".join(self.components),
            negated="NOT " if self.is_negated else "",
            lookup=self.lookup,
            value=self.value,
            callable=callable_repr,
        )

    def __eq__(self, other):
        return hash(self) == hash(other)

    def __hash__(self):
        return hash(repr(self))


class LookupConfig(object):
    """
    Lookup configuration which is used by :class:`.FilterSet`
    to create a :class:`.FilterSpec`.

    The main purpose of this config is to allow the use
    if recursion in :class:`.FilterSet`. Each lookup key
    (the keys in the querystring) is parsed into
    a nested one-key dictionary which lookup config stores.

    For example the querystring::

        ?user__profile__email__endswith=gmail.com

    is parsed into the following config::

        {
            'user': {
                'profile': {
                    'email': {
                        'endswith': 'gmail.com'
                    }
                }
            }
        }

    Attributes
    ----------
    key : str
        Full lookup key from the querystring.
        For example ``user__profile__email__endswith``
    data : dict, str
        Either:

        * nested dictionary where the key is the next key within
          the lookup chain and value is another :class:`.LookupConfig`
        * the filtering value as provided in the querystring value

    Parameters
    ----------
    key : str
        Full lookup key from the querystring.
    data : dict, str
        A regular vanilla Python dictionary.
        This class automatically converts nested
        dictionaries to instances of :class:`.LookupConfig`.
        Alternatively a filtering value as provided
        in the querystring.
    """

    def __init__(self, key, data):
        if isinstance(data, dict):
            data = {k: self.__class__(key, v) for k, v in data.items()}

        self.key = key
        self.data = data

    def is_key_value(self):
        """
        Check if this :class:`.LookupConfig` is not a nested :class:`.LookupConfig`
        but instead the value is a non-dict value.
        """
        return len(self.data) == 1 and not isinstance(self.value, dict)

    @property
    def name(self):
        """
        If the ``data`` is nested :class:`.LookupConfig`,
        this gets its first lookup key.
        """
        return next(iter(self.data.keys()))

    @property
    def value(self):
        """
        If the ``data`` is nested :class:`.LookupConfig`,
        this gets its first lookup value which could either
        be another :class:`.LookupConfig` or actual filtering value.
        """
        return next(iter(self.data.values()))

    def as_dict(self):
        """
        Converts the nested :class:`.LookupConfig` to a regular ``dict``.
        """
        if isinstance(self.data, dict):
            return {k: v.as_dict() for k, v in self.data.items()}
        return self.data

    def __repr__(self):
        return "<{} {}=>{}>".format(
            self.__class__.__name__, self.key, repr(self.as_dict())
        )


class SubClassDict(dict):
    """
    Special-purpose ``dict`` with special getter for looking up
    values by finding matching subclasses.

    This is better illustrated in an example::

        >>> class Klass(object): pass
        >>> class Foo(object): pass
        >>> class Bar(Foo): pass
        >>> mapping = SubClassDict({
        ...     Foo: 'foo',
        ...     Klass: 'klass',
        ... })
        >>> print(mapping.get(Klass))
        klass
        >>> print(mapping.get(Foo))
        foo
        >>> print(mapping.get(Bar))
        foo
    """

    def get(self, k, d=None):
        """
        If no value is found by using Python's default implementation,
        try to find the value where the key is a base class of the
        provided search class.
        """
        value = super(SubClassDict, self).get(k, d)

        # try to match by value
        if value is d and inspect.isclass(k):
            for klasses, v in self.items():
                if not isinstance(klasses, (list, tuple)):
                    klasses = [klasses]
                for klass in klasses:
                    if inspect.isclass(klass) and issubclass(k, klass):
                        return v

        return value


def dictify(obj):
    """
    Convert any object to a dictionary.

    If the given object is already an instance of a dict,
    it is directly returned. If not, then all the public
    attributes of the object are returned as a dict.
    """
    if isinstance(obj, dict):
        return obj
    else:
        return {k: getattr(obj, k) for k in dir(obj) if not k.startswith("_")}


def dict_pop(key, d):
    """
    Pop key from dictionary and return updated dictionary
    """
    d.pop(key, None)
    return d


@contextmanager
def suppress(e):
    """
    Suppress given exception type

    For example::

        >>> with suppress(ValueError):
        ...     print('test')
        ...     raise ValueError
        test
    """
    try:
        yield
    except e:
        pass



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals
import itertools

from sqlalchemy import false, func
from sqlalchemy.orm import class_mapper
from sqlalchemy.sql.expression import not_

from .base import BaseFilterBackend


__all__ = ["SQLAlchemyFilterBackend"]


def lower(value):
    try:
        return value.lower()
    except AttributeError:
        return value


class SQLAlchemyFilterBackend(BaseFilterBackend):
    """
    Filter backend for filtering SQLAlchemy query objects.

    .. warning::
        The filter backend can **ONLY** filter SQLAlchemy's query objects.
        Passing any other datatype for filtering will kill happy bunnies
        under rainbow.

    .. warning::
        The filter backend can **ONLY** filter query objects which query a single
        entity (e.g. query a single model or model column).
        If query object queries multiple entities, ``AssertionError``
        will be raised.
    """

    name = "sqlalchemy"
    supported_lookups = {
        "contains",
        "endswith",
        "exact",
        "gt",
        "gte",
        "icontains",
        "iendswith",
        "iexact",
        "iin",
        "in",
        "isnull",
        "istartswith",
        "lt",
        "lte",
        "range",
        "startswith",
    }

    def __init__(self, *args, **kwargs):
        super(SQLAlchemyFilterBackend, self).__init__(*args, **kwargs)

        assert len(self.queryset._entities) == 1, (
            "{} does not support filtering when multiple entities "
            "are being queried (e.g. session.query(Foo, Bar))."
            "".format(self.__class__.__name__)
        )

    def empty(self):
        """
        Get empty queryset
        """
        return self.queryset.filter(false())

    def get_model(self):
        """
        Get the model from the given queryset
        """
        return self.queryset._only_entity_zero().mapper.class_

    def filter_by_specs(self, queryset):
        """
        Filter SQLAlchemy query object by applying all filter specifications

        The filtering is done by calling ``filter`` with all appropriate
        filter clauses. Additionally if any filter specifications filter
        by related models, those models are joined as necessary.
        """
        if not self.regular_specs:
            return queryset

        clauses = [self.build_clause(spec) for spec in self.regular_specs]
        conditions, joins = zip(*clauses)
        joins = list(itertools.chain(*joins))

        if joins:
            queryset = queryset.join(*joins)

        return queryset.filter(*conditions)

    def build_clause(self, spec):
        """
        Construct SQLAlchemy binary expression filter clause from the
        given filter specification.

        Parameters
        ----------
        spec : FilterSpec
            Filter specification for which to construct filter clause

        Returns
        -------
        tuple
            Tuple of filter binary expression clause and and a list of
            model attributes/descriptors which should be joined when
            doing filtering. If these attributes are not joined,
            SQLAlchemy will not join appropriate tables hence
            wont be able to correctly filter data.
        """
        to_join = []

        model = self.model
        for component in spec.components:
            _field = getattr(model, component)
            field = self._get_properties_for_model(model)[component]
            try:
                model = self._get_related_model_for_field(field)
            except AttributeError:
                break
            else:
                to_join.append(_field)

        # cant directly compare instrumented attributes
        # so need to collect object ids which are unique
        # since they are model singletons
        existing_eagerloads_ids = []
        for option in self.queryset._with_options:
            for suboption in itertools.takewhile(
                lambda i: i.strategy[0] == ("lazy", "joined"), option._to_bind
            ):
                existing_eagerloads_ids.append([id(i) for i in suboption.path])

        already_joined_ids = set()
        for i in range(1, len(to_join) + 1):
            sub_to_join_ids = [id(i) for i in to_join[:i]]
            if sub_to_join_ids in existing_eagerloads_ids:
                already_joined_ids |= set(sub_to_join_ids)

        n_already_joined = len(already_joined_ids)
        to_join = to_join[n_already_joined:]

        builder = getattr(self, "_build_clause_{}".format(spec.lookup))
        column = self._get_attribute_for_field(field)
        clause = builder(spec, column)

        if spec.is_negated:
            clause = not_(clause)

        return clause, to_join

    def _build_clause_contains(self, spec, column):
        return column.contains(spec.value)

    def _build_clause_endswith(self, spec, column):
        return column.endswith(spec.value)

    def _build_clause_exact(self, spec, column):
        return column == spec.value

    def _build_clause_gt(self, spec, column):
        return column > spec.value

    def _build_clause_gte(self, spec, column):
        return column >= spec.value

    def _build_clause_icontains(self, spec, column):
        return func.lower(column).contains(lower(spec.value))

    def _build_clause_iendswith(self, spec, column):
        return func.lower(column).endswith(lower(spec.value))

    def _build_clause_iexact(self, spec, column):
        return func.lower(column) == lower(spec.value)

    def _build_clause_in(self, spec, column):
        return column.in_(spec.value)

    def _build_clause_iin(self, spec, column):
        return func.lower(column).in_(lower(i) for i in spec.value)

    def _build_clause_isnull(self, spec, column):
        if spec.value:
            return column == None  # noqa
        else:
            return column != None  # noqa

    def _build_clause_istartswith(self, spec, column):
        return func.lower(column).startswith(lower(spec.value))

    def _build_clause_lt(self, spec, column):
        return column < spec.value

    def _build_clause_lte(self, spec, column):
        return column <= spec.value

    def _build_clause_range(self, spec, column):
        return column.between(*spec.value)

    def _build_clause_startswith(self, spec, column):
        return column.startswith(spec.value)

    @classmethod
    def _get_properties_for_model(cls, model):
        """
        Get column properties dict for the given model where
        keys are field names and values are column properties
        (e.g. ``ColumnProperty``) or related classes.
        """
        mapper = class_mapper(model)
        return {i.key: i for i in mapper.iterate_properties}

    @classmethod
    def _get_column_for_field(cls, field):
        """
        Get a ``Column`` instance from the model property instance
        (e.g. ``ColumnProperty`` class or related)
        """
        return field.columns[0]

    @classmethod
    def _get_attribute_for_field(cls, field):
        """
        Get the model class attribute/descriptor from property instance
        (e.g. ``ColumnProperty`` class or related)
        """
        return field.class_attribute

    @classmethod
    def _get_related_model_for_field(cls, field):
        """
        Get related model to which field has relationship to
        from property instance (e.g. ``ColumnProperty`` class or related)
        """
        return field._dependency_processor.mapper.class_



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

import coreapi
import coreschema
from django import forms
from django.db.models.constants import LOOKUP_SEP

from ..fields import MultipleValuesField
from ..filtersets import FilterSet
from .drf import URLFilterBackend


DESCRIPTION = {
    "contains": "Match when string contains given substring",
    "day": "Match by day of the month",
    "endswith": "Match when string ends with given substring",
    "exact": "Match exactly the value as is",
    "gt": "Match when value is greater then given value",
    "gte": "Match when value is greater or equal then given value",
    "hour": "Match by the hour (24 hour) value of the timestamp",
    "icontains": "Case insensitive match when string contains given substring",
    "iendswith": "Case insensitive match when string ends with given substring",
    "iexact": "Case insensitive match exactly the value as is",
    "iin": "Case insensitive match when value is any of given comma separated values",
    "in": "Match when value is any of given comma separated values",
    "iregex": "Case insensitive match string by regex pattern",
    "isnull": "Match when value is NULL",
    "istartswith": "Case insensitive match when string starts with given substring",
    "lt": "Match when value is less then given value",
    "lte": "Match when value is less or equal then given value",
    "minute": "Match by the minute value of the timestamp",
    "month": "Match by the month value of the timestamp",
    "range": "Match when value is within comma separated range",
    "regex": "Match string by regex pattern",
    "second": "Match by the second value of the timestamp",
    "startswith": "Match when string starts with given substring",
    "week_day": "Match by week day (1-Sunday to 7-Saturday) of the timestamp",
    "year": "Match by the year value of the timestamp",
}

FORM_FIELD_TO_SCHEMA = {
    MultipleValuesField: coreschema.Array,
    forms.BooleanField: coreschema.Boolean,
    forms.IntegerField: coreschema.Integer,
    forms.FloatField: coreschema.Number,
}


def _field_to_schema(field, lookup):
    form_field = field.get_form_field(lookup)
    return FORM_FIELD_TO_SCHEMA.get(type(form_field), coreschema.String)(
        description=DESCRIPTION.get(lookup)
    )


def _all_filters(filterset, prefix=()):
    for name, field in filterset.filters.items():
        if isinstance(field, FilterSet):
            for i in _all_filters(field, prefix=prefix + (name,)):
                yield i

        else:
            yield coreapi.Field(
                name=LOOKUP_SEP.join(prefix + (name,)),
                required=False,
                location="query",
                schema=_field_to_schema(field, field.default_lookup),
            )
            if field.no_lookup:
                continue
            for lookup in field.lookups:
                yield coreapi.Field(
                    name=LOOKUP_SEP.join(prefix + (name, lookup)),
                    required=False,
                    location="query",
                    schema=_field_to_schema(field, lookup),
                )


class CoreAPIURLFilterBackend(URLFilterBackend):
    """
    Same as :py:class:`url_filter.integrations.drf.DjangoFilterBackend` except
    this backend also implements coreapi interface for autogenerated API docs.
    """

    def get_schema_fields(self, view):
        """
        Get coreapi filter definitions

        Returns all filters including their supported lookups.
        """
        queryset = view.get_queryset()
        filter_class = self.get_filter_class(view, queryset)

        return (
            []
            if not filter_class
            else list(_all_filters(filter_class(data={}, queryset=queryset)))
        )



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals
import abc
import re
from functools import wraps

import six
from cached_property import cached_property
from django import forms
from django.core.exceptions import ValidationError

from .constants import StrictMode
from .fields import MultipleValuesField
from .utils import FilterSpec, dict_pop


MANY_LOOKUP_FIELD_OVERWRITES = {
    "in": lambda **kwargs: MultipleValuesField(min_values=1, **kwargs),
    "iin": lambda **kwargs: MultipleValuesField(min_values=1, **kwargs),
    "range": lambda **kwargs: MultipleValuesField(
        min_values=2, max_values=2, **dict_pop("all_valid", kwargs)
    ),
}

LOOKUP_FIELD_OVERWRITES = {
    "isnull": forms.BooleanField(required=False),
    "second": forms.IntegerField(min_value=0, max_value=59),
    "minute": forms.IntegerField(min_value=0, max_value=59),
    "hour": forms.IntegerField(min_value=0, max_value=23),
    "week_day": forms.IntegerField(min_value=1, max_value=7),
    "day": forms.IntegerField(min_value=1, max_value=31),
    "month": forms.IntegerField(),
    "year": forms.IntegerField(min_value=0, max_value=9999),
    "regex": forms.CharField(),
    "iregex": forms.CharField(),
}

LOOKUP_CALLABLE_FROM_METHOD_REGEX = re.compile(
    r"^filter_(?P<filter>[\w\d]+)_for_(?P<backend>[\w\d]+)$"
)


class BaseFilter(six.with_metaclass(abc.ABCMeta, object)):
    """
    Base class to be used for defining both filters and filtersets.

    This class implements the bare-minimum functions which are used across
    both filters and filtersets however all other functionality must be
    implemented in subclasses. Additionally by using a single base class,
    both filters and filtersets inherit from the same base class hence
    instance checks can be easily done by filteset's metaclass in order
    to find all declared filters defined in it.

    Parameters
    ----------
    source : str
        Name of the attribute for which which filter applies to
        within the model of the queryset to be filtered
        as given to the :class:`.FilterSet`.

    Attributes
    ----------
    parent : :class:`.FilterSet`
        Parent :class:`.FilterSet` to which this filter is bound to
    name : str
        Name of the field as it is defined in parent :class:`.FilterSet`
    is_bound : bool
        If this filter has been bound to a parent yet
    """

    def __init__(self, source=None, *args, **kwargs):
        self._source = source
        self.parent = None
        self.name = None
        self.is_bound = False

    def __repr__(self):
        data = self.repr()
        data = data if six.PY3 else data.encode("utf-8")
        return data

    @abc.abstractmethod
    def repr(self, prefix=""):
        """
        Get the representation of the filter or its subclasses.

        Subclasses **must** overwrite this method.

        .. note::
            This class should return unicode text data

        Parameters
        ----------
        prefix : str
            All nested filtersets provide useful representation of the complete
            filterset including all descendants however in that case descendants
            need to be indented in order for the representation to get structure.
            This parameter is used to do just that. It specifies the prefix
            the representation must use before returning any of its representations.
        """

    @property
    def source(self):
        """
        Source field/attribute in queryset model to be used for filtering.

        This property is helpful when ``source`` parameter is not provided
        when instantiating :class:`.BaseFilter` or its subclasses since it will
        use the filter name as it is defined in the :class:`.FilterSet`.
        For example::

            >>> from .filtersets import FilterSet
            >>> class MyFilterSet(FilterSet):
            ...     foo = Filter(form_field=forms.CharField())
            ...     bar = Filter(source='stuff', form_field=forms.CharField())
            >>> fs = MyFilterSet()
            >>> print(fs.filters['foo'].source)
            foo
            >>> print(fs.filters['bar'].source)
            stuff
        """
        return self._source or self.name

    @property
    def components(self):
        """
        List of all components (source names) of all parent filtersets.
        """
        if self.parent is None:
            return []
        return self.parent.components + [self.source]

    def bind(self, name, parent):
        """
        Bind the filter to the filterset.

        This method should be used by the parent :class:`.FilterSet`
        since it allows to specify the parent and name of each
        filter within the filterset.
        """
        self.name = name
        self.parent = parent
        self.is_bound = True

    @property
    def root(self):
        """
        This gets the root filterset.
        """
        if self.parent is None:
            return self
        return self.parent.root


class Filter(BaseFilter):
    """
    Class which job is to convert leaf :class:`.LookupConfig` to
    :class:`.FilterSpec`

    Each filter by itself is meant to be used a "filter" (field) in the
    :class:`.FilterSet`.

    Examples
    --------

    ::

        >>> from .filtersets import FilterSet
        >>> class MyFilterSet(FilterSet):
        ...     foo = Filter(forms.CharField())
        ...     bar = Filter(forms.IntegerField())

    Parameters
    ----------
    form_field : Field
        Instance of Django's ``forms.Field`` which will be used
        to clean the filter value as provided in the queryset.
        For example if field is ``IntegerField``, this filter
        will make sure to convert the filtering value to integer
        before creating a :class:`.FilterSpec`.
    lookups : list, optional
        List of strings of allowed lookups for this filter.
        By default all supported lookups are allowed.
    default_lookup : str, optional
        If the lookup is not provided in the querystring lookup key,
        this lookup will be used. By default ``exact`` lookup is used.
        For example the default lookup is used when querystring key is
        ``user__profile__email`` which is missing the lookup so ``exact``
        will be used.
    is_default : bool, optional
        Boolean specifying if this filter should be used as a default
        filter in the parent :class:`.FilterSet`.
        By default it is ``False``.
        Primarily this is used when querystring lookup key
        refers to a nested :class:`.FilterSet` however it does not specify
        which filter to use. For example lookup key ``user__profile``
        intends to filter something in the user's profile however
        it does not specify by which field to filter on.
        In that case the default filter within profile :class:`.FilterSet`
        will be used. At most, one default filter should be provided
        in the :class:`.FilterSet`.
    no_lookup : bool, optional
        When ``True``, this filter does not allow to explicitly specify
        lookups in the URL. For example ``id__gt`` will not be allowed.
        This is useful when a given filter should only support a single
        lookup and that lookup name should not be exposed in the URL.
        This is of particular use when defining custom callable filters.
        By default it is ``False``.

    Attributes
    ----------
    form_field : Field
        Django form field which is provided in initialization which
        should be used to validate data as provided in the querystring
    default_lookup : str
        Default lookup to be used as provided in initialization
    is_default : bool
        If this filter should be a default filter as provided in initialization
    no_lookup : str
        If this filter should not support explicit lookups as provided in initialization
    """

    def __init__(
        self,
        form_field,
        lookups=None,
        default_lookup="exact",
        is_default=False,
        no_lookup=False,
        *args,
        **kwargs
    ):
        super(Filter, self).__init__(*args, **kwargs)
        self.form_field = form_field
        self._given_lookups = lookups
        self.default_lookup = default_lookup or self.default_lookup
        self.is_default = is_default
        self.no_lookup = no_lookup

    def repr(self, prefix=""):
        """
        Get custom representation of the filter

        The representation includes the following information:

        * filter class name
        * source name (same as :attr:`.source`) when filter is bound to parent
        * primary form field (same as :attr:`.form_field`)
        * which lookups this filter supports
        * default lookup (same as :attr:`.default_lookup`)
        * if the filter is a default filter (same as :attr:`.is_default`) when
          filter is bound to parent
        * if this filter does not support explicit lookups (same as :attr:`.no_lookup`)
        """
        return (
            "{name}("
            "{source}"
            "form_field={form_field}, "
            "lookups={lookups}, "
            'default_lookup="{default_lookup}", '
            "{is_default}"
            "no_lookup={no_lookup}"
            ")"
            "".format(
                name=self.__class__.__name__,
                source='source="{}", '.format(self.source) if self.is_bound else "",
                form_field=self.form_field.__class__.__name__,
                lookups=self._given_lookups or "ALL",
                default_lookup=self.default_lookup,
                is_default="is_default={}, ".format(self.is_default)
                if self.is_bound
                else "",
                no_lookup=self.no_lookup,
            )
        )

    @cached_property
    def lookups(self):
        """
        Cached property for getting lookups this filter supports

        The reason why we need as a property is because lookups
        cant be hardcoded. There are 3 main distinct possibilities
        which drive which lookups are supported:

        * lookups were explicitly provided in the filter instantiation
          in which case we use those lookups. For example::

              >>> f = Filter(forms.CharField(), lookups=['exact', 'contains'])
        * when filter is already bound to a parent filterset and root
          filterset has a defined ``filter_backend`` we use supported
          lookups as explicitly defined by the backend. This is necessary
          since different backends support different sets of lookups.
        * when nether lookups are explicitly provided and filter is not bound
          yet we have no choice but not support any lookups and so we
          use empty set as supported lookups
        """
        if self._given_lookups:
            return set(self._given_lookups)
        if hasattr(self.root, "filter_backend"):
            return self.root.filter_backend.supported_lookups
        return set()

    def get_form_field(self, lookup):
        """
        Get the form field for a particular lookup.

        This method does not blindly return :attr:`.form_field` attribute
        since some lookups require to use different validations.
        For example if the :attr:`.form_field` is ``CharField`` but
        the lookup is ``isnull``, it makes more sense to use
        ``BooleanField`` as form field.

        Parameters
        ----------
        lookup : str
            Name of the lookup

        Returns
        -------
        Field
            Instantiated form field appropriate for the given lookup.
        """
        if lookup in MANY_LOOKUP_FIELD_OVERWRITES:
            return MANY_LOOKUP_FIELD_OVERWRITES[lookup](
                child=self.form_field,
                all_valid=getattr(self.root, "strict_mode", StrictMode.fail)
                == StrictMode.fail,
            )
        elif lookup in LOOKUP_FIELD_OVERWRITES:
            return LOOKUP_FIELD_OVERWRITES[lookup]
        else:
            return self.form_field

    def clean_value(self, value, lookup):
        """
        Clean the filter value as appropriate for the given lookup.

        Parameters
        ----------
        value : str
            Filter value as given in the querystring to be validated
            and cleaned by using appropriate Django form field
        lookup : str
            Name of the lookup

        See Also
        --------
        get_form_field
        """
        form_field = self.get_form_field(lookup)
        return form_field.clean(value)

    def get_spec(self, config):
        """
        Get the ``FilterSpec`` for the provided ``config``.

        Parameters
        ----------
        config : LookupConfig
            Lookup configuration for which to build ``FilterSpec``.
            The lookup should be a leaf configuration otherwise
            ``ValidationError`` is raised.

        Returns
        -------
        FilterSpec
            spec constructed from the given configuration.
        """
        # lookup was explicitly provided
        if isinstance(config.data, dict):
            if not config.is_key_value():
                raise ValidationError(
                    "Invalid filtering data provided. "
                    "Data is more complex then expected. "
                    "Most likely additional lookup was specified "
                    "after the final lookup (e.g. field__in__equal=value)."
                )

            if self.no_lookup:
                raise ValidationError(
                    "Lookup was explicit used in filter specification. "
                    "This filter does not allow to specify lookup."
                )

            lookup = config.name
            value = config.value.data

        # use default lookup
        else:
            lookup = self.default_lookup
            value = config.data

        if lookup not in self.lookups:
            raise ValidationError('"{}" lookup is not supported'.format(lookup))

        is_negated = "!" in config.key
        value = self.clean_value(value, lookup)

        return FilterSpec(self.components, lookup, value, is_negated)


def form_field_for_filter(form_field):
    """
    Decorator for specifying form field for a custom callable filter
    on the filter callable method

    Examples
    --------
    ::

        >>> class MyFilterCallable(CallableFilter):
        ...     @form_field_for_filter(forms.CharField())
        ...     def filter_foo_for_django(self, queryset, spec):
        ...         return queryset

    Parameters
    ----------
    form_field : Field
        Django form field which should be used for the decorated
        custom callable filter

    Returns
    -------
    func
        Function which can be used to decorate a class method
    """

    def wrapper(f):
        @wraps(f)
        def inner(self, *args, **kwargs):
            return f(self, *args, **kwargs)

        inner.form_field = form_field

        return inner

    return wrapper


class CallableFilter(Filter):
    """
    Custom filter class meant to be subclassed in order to add
    support for custom lookups via custom callables

    The goal of this filter is to provide:

    * support for custom callbacks (or overwrite existing ones)
    * support different filtering backends

    Custom callable functions for lookups and different backends
    are defined via class methods by using the following method
    name pattern::

        filter_<lookup_name>_for_<backend_name>

    Obviously multiple methods can be used to implement functionality
    for multiple lookups and/or backends. This makes callable filters
    pretty flexible and ideal for implementing custom reusable filtering
    filters which follow DRY.

    Examples
    --------
    ::

        >>> from django.http import QueryDict
        >>> from .filtersets import FilterSet

        >>> class MyCallableFilter(CallableFilter):
        ...     @form_field_for_filter(forms.CharField())
        ...     def filter_foo_for_django(self, queryset, spec):
        ...         f = queryset.filter if not spec.is_negated else queryset.exclude
        ...         return f(foo=spec.value)
        ...     def filter_foo_for_sqlalchemy(self, queryset, spec):
        ...         op = operator.eq if not spec.is_negated else operator.ne
        ...         return queryset.filter(op(Foo.foo, spec.value))

        >>> class MyFilterSet(FilterSet):
        ...     field = MyCallableFilter()

        >>> f = MyFilterSet(data=QueryDict('field__foo=bar'), queryset=[])

    .. note::
        Unlike base class :class:`.Filter` this filter makes
        ``form_field`` parameter optional. Please note however that
        when ``form_field`` parameter is not provided, all custom
        filter callables should define their own appropriate form fields
        by using :func:`.form_field_for_filter`.
    """

    def __init__(self, form_field=None, *args, **kwargs):
        # need to overwrite to make form_field optional
        super(CallableFilter, self).__init__(form_field, *args, **kwargs)

    @cached_property
    def lookups(self):
        """
        Get all supported lookups for the filter

        This property is identical to the super implementation except it also
        finds all custom lookups from the class methods and adds them to the
        set of supported lookups as returned by the super implementation.
        """
        lookups = super(CallableFilter, self).lookups

        r = LOOKUP_CALLABLE_FROM_METHOD_REGEX
        custom_lookups = {
            m.groupdict()["filter"]
            for m in (r.match(i) for i in dir(self))
            if m and m.groupdict()["backend"] == self.root.filter_backend.name
        }

        return lookups | custom_lookups

    def _get_filter_method_for_lookup(self, lookup):
        name = "filter_{}_for_{}".format(lookup, self.root.filter_backend.name)
        return getattr(self, name)

    def get_form_field(self, lookup):
        """
        Get the form field for a particular lookup.

        This method attempts to return form field for custom callables
        as set by :func:`.form_field_for_filter`. When either custom
        lookup is not set or its form field is not set, super implementation
        is used to get the form field. If form field at that point is not
        found, this method raises ``AssertionError``. That can only happen
        when `form_field` parameter is not provided during initialization.
        """
        try:
            return self._get_filter_method_for_lookup(lookup).form_field
        except AttributeError:
            pass

        form_field = super(CallableFilter, self).get_form_field(lookup)

        assert form_field is not None, (
            "{name} was not provided form_field parameter in initialization "
            "(e.g. {name}(form_field=CharField)) and form_field was not "
            "provided for the lookup. If the lookup is a custom filter callable "
            "you should provide form_field by using @form_field_for_filter "
            "decorator. If the lookup is a normal lookup, then please either "
            "provide form_field parameter or overwrite get_form_field()."
            "".format(name=self.__class__.__name__)
        )

        return form_field

    def get_spec(self, config):
        """
        Get the :class:`.FilterSpec` for the given :class:`.LookupConfig`
        with appropriately set :attr:`.FilterSpec.filter_callable`
        when the lookup is a custom lookup
        """
        spec = super(CallableFilter, self).get_spec(config)
        spec.filter_callable = self._get_filter_method_for_lookup(spec.lookup)
        return spec



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function
import os

from setuptools import find_packages, setup

from url_filter import __author__, __version__


def read(fname):
    with open(os.path.join(os.path.dirname(__file__), fname), "rb") as fid:
        return fid.read().decode("utf-8")


authors = read("AUTHORS.rst")
history = read("HISTORY.rst").replace(".. :changelog:", "")
licence = read("LICENSE.rst")
readme = read("README.rst")

req = read("requirements.txt").splitlines()
dev_req = read("requirements-dev.txt").splitlines()[2:]

requirements = req + ["setuptools"]
test_requirements = req + dev_req

setup(
    name="django-url-filter",
    version=__version__,
    author=__author__,
    description="Django URL Filter provides a safe way to filter data via human-friendly URLs.",
    long_description="\n\n".join([readme, history, authors, licence]),
    url="https://github.com/miki725/django-url-filter",
    license="MIT",
    packages=find_packages(exclude=["test_project*", "tests*"]),
    install_requires=requirements,
    test_suite="tests",
    tests_require=test_requirements,
    keywords=" ".join(["django django-rest-framework"]),
    classifiers=[
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Development Status :: 2 - Pre-Alpha",
    ],
)
